;; hack-quail.el -- correct some bugs and limitations of quail

;; Copyright (C) 1998 Milan Zamazal <pdm@fi.muni.cz>

;; Author:   Milan Zamazal <pdm@fi.muni.cz>
;; Version:  4.0
;; Keywords: quail, i18n, Czech, Slovak

;; COPYRIGHT NOTICE
;;
;; This program is free software; you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by the Free
;; Software Foundation, version 2 of the License.
;;
;; This program is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
;; or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
;; for more details.
;;
;; You can find the GNU General Public License at
;; http://www.gnu.org/copyleft/gpl.html
;; or you can write to the Free Software Foundation, Inc., 675 Mass Ave.,
;; Cambridge, MA 02139, USA.

;;; Installation:
;;
;; To install this file, just put it to some of load-path directories and add
;;  (require 'hack-quail)
;; to your ~/.emacs.

;;; Commentary:
;;
;; This file tries to remove some quail problems without affecting the original
;; file.  These problems are:
;; - Bug: quail doesn't allow to use letter bindings of current modes.
;; - Limitation: numpad keys are equal to digit keys for quail.
;; - Limitation: Input method doesn't allow non-ASCII keys.
;; This hack tries to remove these problems, but only with respect to Emacs
;; Czech and Slovak keyboard.  If you need anything else, you should probably
;; try to contact the Quail maintainer firstly.

;;; Code:

(require 'quail)


;;; Bug: quail doesn't allow to use letter bindings of current modes.

;; Solution: We redefine the `quail-start-translation-function.  Our wrapper
;; firstly looks through some currently active keytables and if it finds some
;; definition for the given key, it uses it.

(defun hack-quail-is-special ()
  "Return last key definition according to current modes."
  ;; `minor-mode-key-binding' doesn't work in 20.2
  (when (numberp last-command-event)
    (let ((key (char-to-string last-command-event))
	  (keymaps (current-minor-mode-maps))
	  command
	  is-special)
      (while (and (not is-special)
		  keymaps)
	(setq command (lookup-key (car keymaps) key))
	(if (memq command '(nil quail-start-translation self-insert-command))
	    (setq keymaps (cdr keymaps))
	  (setq is-special command)))
      (or is-special
	  (local-key-binding key)))))

;; Redefine quail-start-translation
(unless (featurep 'hack-quail)
  (fset 'hack-quail-start-translation
	(symbol-function 'quail-start-translation)))
(defun quail-start-translation (arg)
  (interactive "p")
  (if (hack-quail-is-special)
      (progn
	(setq unread-command-events (list last-command-event))
	(setq quail-mode nil)
	(add-hook 'pre-command-hook 'hack-quail-mode-reset))
    (call-interactively 'hack-quail-start-translation)))


;;; Limitation: numpad keys are equal to digit keys for quail.
;;; (This is very annoying especially when using Czech or Slovak keyboard.)

;; Solution: We define these keys in `quail-mode-map'.

(define-key quail-mode-map [kp-1] 'quail-start-translation)
(define-key quail-mode-map [kp-2] 'quail-start-translation)
(define-key quail-mode-map [kp-3] 'quail-start-translation)
(define-key quail-mode-map [kp-4] 'quail-start-translation)
(define-key quail-mode-map [kp-5] 'quail-start-translation)
(define-key quail-mode-map [kp-6] 'quail-start-translation)
(define-key quail-mode-map [kp-7] 'quail-start-translation)
(define-key quail-mode-map [kp-8] 'quail-start-translation)
(define-key quail-mode-map [kp-9] 'quail-start-translation)
(define-key quail-mode-map [kp-0] 'quail-start-translation)
(define-key quail-mode-map [kp-add] 'quail-start-translation)


;;; Limitation: Input method doesn't allow non-ASCII keys.
;; (So non-ASCII dead keys cannot be used.)

;; Solution: We put into key-translation-map functions, which translate some
;; keys and dead key according to the current input method.  The real dead key
;; is represented by some ASCII key in leim keyboard definitions.  This ASCII
;; key disables quail mode temporarily, so that it could be inserted not as a
;; dead key.
;; Terrible, but necessary.


;; User configuration

(defvar hack-quail-dead-key nil
  "*The key dedicated to be the dead-key on the \"czech-prog\" keyboard.
If it should be `+', set this variable to nil.")

(defvar hack-quail-hacked-key "+"
  "The substituted dead key from leim keyboard definitions.
This must be string.
This should be set just once, before `hack-quail' is loaded.")

(defvar hack-quail-hack-methods '("czech-prog-1" "czech-prog-2"
				  "slovak-prog-1" "slovak-prog-3")
  "*List of keyboard input methods affected by the plus translation hack.")


;; Auxiliary functions, etc.

(defvar hack-quail-dead-flag nil)

(defun hack-quail-translate-plus (arg)
  "Translate `hack-quail-hacked-key' key according to the current-input-method."
  (add-hook 'pre-command-hook 'hack-quail-dead-flag-reset)
  (when (and (member current-input-method hack-quail-hack-methods)
	     (not hack-quail-dead-flag))
    (add-hook 'pre-command-hook 'hack-quail-mode-reset)
    (setq quail-mode nil))
  hack-quail-hacked-key)

(defun hack-quail-mode-reset ()
  "Remove `hack-quail-mode-reset' from all hooks."
  (remove-hook 'pre-command-hook 'hack-quail-mode-reset)
  (setq quail-mode t))

(defun hack-quail-dead-flag-reset ()
  "Remove `hack-quail-dead-flag-reset' from all hooks."
  (remove-hook 'pre-command-hook 'hack-quail-dead-flag-reset)
  (remove-hook 'input-method-after-insert-chunk-hook
	       'hack-quail-dead-flag-reset)
  (setq hack-quail-dead-flag nil))

(defun hack-quail-translate-dead-key (arg)
  "Translate dead and/or substituted key."
  (when quail-mode
    (setq hack-quail-dead-flag t)
    (add-hook 'input-method-after-insert-chunk-hook
	      'hack-quail-dead-flag-reset))
  (if hack-quail-dead-flag hack-quail-hacked-key hack-quail-dead-key))

;; Initialize key translation map
(when (and hack-quail-dead-key
	   (or (not (stringp hack-quail-dead-key))
	       (not (string= hack-quail-dead-key hack-quail-hacked-key))))
  (or key-translation-map
      (setq key-translation-map (make-sparse-keymap)))
  (define-key key-translation-map
    hack-quail-dead-key 'hack-quail-translate-dead-key)
  (define-key key-translation-map
    hack-quail-hacked-key 'hack-quail-translate-plus))


;;; Announce

(provide 'hack-quail)
