;;; string-recode.el - recoding of sequences

;; Copyright (C) 1997, 1998 Milan Zamazal

;; Author:     Milan Zamazal <pdm@fi.muni.cz>
;; Version:    4.0
;; Keywords:   encodings, conversion, Czech

;; COPYRIGHT NOTICE
;;
;; This program is free software; you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by the Free
;; Software Foundation, version 2 of the License.
;;
;; This program is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
;; or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
;; for more details.
;;
;; You can find the GNU General Public License at
;; http://www.gnu.org/copyleft/gpl.html
;; or you can write to the Free Software Foundation, Inc., 675 Mass Ave.,
;; Cambridge, MA 02139, USA.

;;; Installation:

;; Put this file somewhere into load-path and to your ~/.emacs the following
;; lines:
;;
;;   (autoload 'sr-recode-to-many "string-recode" t)
;;   (autoload 'sr-recode-from-many "string-recode" t)
;;
;; It may also be a good idea to append:
;;
;;   (add-hook 'sgml-mode-hook (lambda () (setq sr-default-table "SGML")))

;;; Commentary:

;; This program performs recoding between strings in the text.  Its primary
;; usage is conversion between Czech accented characters and appropriate
;; (La)TeX sequences.
;;
;; Through the variable `sr-tables' you can specify one input string and one or
;; more output strings.  If more than one output string is specified, then for
;; input-output conversion the first one is used.  For the output-input
;; conversion, the most often appearing sequence in the corrected text is
;; supposed to be the right one and this is the only converted for each given
;; input string.
;;
;; Currently supported conversions are TeX sequences, SGML sequences, and
;; digraphs (the latter is mostly joke :-).

;;; Code:


(defvar sr-default-table "TeX"
  "*Default recoding table.
See `sr-tables'.")
(make-variable-buffer-local 'sr-default-table)

(defvar sr-tables
  '(("TeX" ""
     (",Ba(B" "\\'a" "\\'{a}")
     (",Bd(B" "\\\"a" "\\\"{a}")
     (",Bh(B" "\\v c" "\\v{c}")
     (",Bo(B" "\\v d" "\\v{d}")
     (",Bi(B" "\\'e" "\\'{e}")
     (",Bl(B" "\\v e" "\\v{e}")
     (",Bm(B" "\\'\\i{}" "\\'{\\i}")
     (",Be(B" "\\'l" "\\'{l}")
     (",B5(B" "\\v l" "\\v{l}")
     (",Br(B" "\\v n" "\\v{n}")
     (",Bs(B" "\\'o" "\\{'o}")
     (",Bv(B" "\\\"o" "\\\"{o}")
     (",Bt(B" "\\^o" "\\^{o}")
     (",B`(B" "\\'r" "\\'{r}")
     (",Bx(B" "\\v r" "\\v{r}")
     (",B9(B" "\\v s" "\\v{s}")
     (",B;(B" "\\v t" "\\v{t}")
     (",Bz(B" "\\'u" "\\'{u}")
     (",By(B" "\\accent23u" "\\r{u}")
     (",B|(B" "\\\"u" "\\\"{u}")
     (",B}(B" "\\'y" "\\'{y}")
     (",B>(B" "\\v z" "\\v{z}")
     (",BA(B" "\\'A" "\\'{A}")
     (",BD(B" "\\\"A" "\\\"{A}")
     (",BH(B" "\\v C" "\\v{C}")
     (",BO(B" "\\v D" "\\v{D}")
     (",BI(B" "\\'E" "\\'{E}")
     (",BL(B" "\\v E" "\\v{E}")
     (",BM(B" "\\'\\I{}" "\\'{\\I}")
     (",BE(B" "\\'L" "\\'{L}")
     (",B%(B" "\\v L" "\\v{L}")
     (",BR(B" "\\v N" "\\v{N}")
     (",BS(B" "\\'O" "\\{'O}")
     (",BV(B" "\\\"O" "\\\"{O}")
     (",BT(B" "\\^O" "\\^{O}")
     (",B@(B" "\\'R" "\\'{R}")
     (",BX(B" "\\v R" "\\v{R}")
     (",B)(B" "\\v S" "\\v{S}")
     (",B+(B" "\\v T" "\\v{T}")
     (",BZ(B" "\\'U" "\\'{U}")
     (",BY(B" "\\accent23U" "\\r{U}")
     (",B\(B" "\\\"U" "\\\"{U}")
     (",B](B" "\\'Y" "\\'{Y}")
     (",B.(B" "\\v Z" "\\v{Z}"))
    ("SGML" ""
     (",Ba(B" "&aacute;")
     (",Bd(B" "&auml;")
     (",Bh(B" "&ccaron;")
     (",Bo(B" "&dcaron;")
     (",Bi(B" "&eacute;")
     (",Bl(B" "&ecaron;")
     (",Bm(B" "&iacute;")
     (",Be(B" "&lacute;")
     (",B5(B" "&lcaron;")
     (",Br(B" "&ncaron;")
     (",Bs(B" "&oacute;")
     (",Bv(B" "&ouml;")
     (",Bt(B" "&ocirc;")
     (",B`(B" "&rcaron;")
     (",Bx(B" "&racute;")
     (",B9(B" "&scaron;")
     (",B;(B" "&tcaron;")
     (",Bz(B" "&ucaron;")
     (",By(B" "&uring;")
     (",B|(B" "&uuml;")
     (",B}(B" "&yacute;")
     (",B>(B" "&zcaron;")
     (",BA(B" "&Aacute;")
     (",BD(B" "&Auml;")
     (",BH(B" "&Ccaron;")
     (",BO(B" "&Dcaron;")
     (",BI(B" "&Eacute;")
     (",BL(B" "&Ecaron;")
     (",BM(B" "&Iacute;")
     (",BE(B" "&Lacute;")
     (",B%(B" "&Lcaron;")
     (",BR(B" "&Ncaron;")
     (",BS(B" "&Oacute;")
     (",BV(B" "&Ouml;")
     (",BT(B" "&Ocirc;")
     (",B@(B" "&Rcaron;")
     (",BX(B" "&Racute;")
     (",B)(B" "&Scaron;")
     (",B+(B" "&Tcaron;")
     (",BZ(B" "&Ucaron;")
     (",BY(B" "&Uring;")
     (",B\(B" "&Uuml;")
     (",B](B" "&Yacute;")
     (",B.(B" "&Zcaron;"))
    ("czech-digraphs" "|"
     ("d,Bl(B" "dje")
     ("t,Bl(B" "tje")
     ("n,Bl(B" "nje")
     ("b,Bl(B" "bje")
     ("p,Bl(B" "pje")
     ("v,Bl(B" "vje")
     ("m,Bl(B" "mje")
     (",Ba(B"  "aa")
     (",Bi(B"  "ee")
     (",Bm(B"  "ii")
     (",Bs(B"  "oo")
     (",Bz(B"  "uu")
     (",By(B"  "uo" "uu")
     (",B}(B"  "yy")
     (",Bh(B"  "cz" "cj")
     (",Bx(B"  "rz" "rj")
     (",B9(B"  "sz" "sj")
     (",Bo(B"  "dj")
     (",Br(B"  "nj")
     (",B;(B"  "tj")
     (",B>(B"  "zj" "zh" "zz")
     ("D,Bl(B" "Dje")
     ("T,Bl(B" "Tje")
     ("N,Bl(B" "Nje")
     ("B,Bl(B" "Bje")
     ("P,Bl(B" "Pje")
     ("V,Bl(B" "Vje")
     ("M,Bl(B" "Mje")
     (",BA(B"  "Aa")
     (",BI(B"  "Ee")
     (",BM(B"  "Ii")
     (",BS(B"  "Oo")
     (",BZ(B"  "Uu")
     (",BY(B"  "Uo" "Uu")
     (",B](B"  "Yy")
     (",BH(B"  "Cz" "Cj")
     (",BX(B"  "Rz" "Rj")
     (",B)(B"  "Sz" "Sj")
     (",BO(B"  "Dj")
     (",BR(B"  "Nj")
     (",B+(B"  "Tj")
     (",B.(B"  "Zj" "Zh" "Zz")
     ("D,BL(B" "DJE")
     ("T,BL(B" "TJE")
     ("N,BL(B" "NJE")
     ("B,BL(B" "BJE")
     ("P,BL(B" "PJE")
     ("V,BL(B" "VJE")
     ("M,BL(B" "MJE")
     (",BA(B"  "AA")
     (",BI(B"  "EE")
     (",BM(B"  "II")
     (",BS(B"  "OO")
     (",BZ(B"  "UU")
     (",BY(B"  "UO" "UU")
     (",B](B"  "YY")
     (",BH(B"  "CZ" "CJ")
     (",BX(B"  "RZ" "RJ")
     (",B)(B"  "SZ" "SJ")
     (",BO(B"  "DJ")
     (",BR(B"  "NJ")
     (",B+(B"  "TJ")
     (",B.(B"  "ZJ" "ZH" "ZZ")))
  "Recoding tables.

It is an association list which elements are of the form

  (NAME SEPARATOR (PATTERN ENCODED ...) ...)

NAME is a name of the recoding.
SEPARATOR is a string inserted when some original text could be confused with
an encoded text.
PATTERN is a string to be encoded.
The first ENCODED element is the encoded PATTERN.  Optional next ENCODED
elements are non-canonical encodings of the PATTERN used in back conversion.")

;;;###autoload
(defun sr-recode-from-many (beg end name)
  "Recode region to the first elements the item NAME of the `sr-tables'."
  (interactive
   (progn
     (barf-if-buffer-read-only)
     (list
      (region-beginning)
      (region-end)
      (sr-read-name))))
  (let* ((end-marker (copy-marker end))
	 (item (cdr (assoc name sr-tables)))
	 (separator (car item))
	 (table (cdr item))
	 recoding-list
	 regexp)
    (save-excursion
      ;; Build replacement list
      (let (winner
	    winner-count
	    choices
	    count)
	(while table
	  (setq winner (car (cdar table))
		winner-count (sr-regexp-count winner beg end)
		choices (cdr (cdar table)))
	  ;; We choose the most often repeating encoding element as the right
	  ;; one used by the user.
	  (while choices
	    (setq count (sr-regexp-count (car choices) beg end))
	    (if (> count winner-count)
		(setq winner (car choices)
		      winner-count count))
	    (setq choices (cdr choices)))
	  (setq recoding-list (cons (cons winner (caar table))
				    recoding-list)
		regexp (if (null regexp)
			   (regexp-quote winner)
			 (concat regexp "\\|" (regexp-quote winner))))
	  (setq table (cdr table))))
      ;; Recode strings
      (goto-char beg)
      (let ((case-fold-search nil))
	(while (and (<= (point) (marker-position end-marker))
		    (re-search-forward regexp end-marker t))
	  (replace-match (cdr (assoc (match-string 0) recoding-list)) 1)))
      ;; Remove separators
      (unless (string= separator "")
	(goto-char beg)
	(while (and (<= (point) (marker-position end-marker))
		    (search-forward separator end-marker t))
	  (delete-char -1)
	  (forward-char 1))))))

;;;###autoload
(defun sr-recode-to-many (beg end name)
  "Recode region to the second elements of the NAME item of `sr-tables'."
  (interactive
   (progn
     (barf-if-buffer-read-only)
     (list
      (region-beginning) 
      (region-end)
      (sr-read-name))))
  (let* ((end-marker (copy-marker end))
	 (item (cdr (assoc name sr-tables)))
	 (separator (car item))
	 (table (cdr item))
	 (regexp (concat (caar table) "\\|" (car (cdar table)))))
    (save-excursion
      ;; Build regular expression
      (let ((table-copy (cdr table))
	    aux)
	(while table-copy
	  (setq aux (car table-copy))
	  (setq regexp (concat regexp "\\|" (car aux) "\\|" (cadr aux)))
	  (setq table-copy (cdr table-copy))))
      ;; Start searching
      (goto-char beg)
      (let ((case-fold-search nil)
	    string-item)
	(while (and (<= (point) (marker-position end-marker))
		    (re-search-forward regexp end-marker t))
	  (setq string-item (assoc (match-string 0) table))
	  (if string-item
	      ;; Replace by an appropriately encoded string
	      (replace-match (cadr string-item) t)
	    ;; Insert separator
	    (goto-char (1+ (match-beginning 0)))
	    (insert separator)))))))


;;; Auxiliary functions

(defvar sr-names-history nil)

(defun sr-read-name ()
  "Read name of conversion from minibuffer."
  (setq sr-default-table
	(completing-read
	 (format "Conversion type (default %s): " sr-default-table)
	 (let ((tables sr-tables)
	       choices)
	   (while tables
	     (setq choices (cons (list (caar tables)) choices))
	     (setq tables (cdr tables)))
	   choices)
	 nil t "" 'sr-names-history sr-default-table)))

(defun sr-regexp-count (regexp beg end)
  "Number of REGEXP occurences in the region."
  (goto-char beg)
  (let ((n 0))
    (while (search-forward regexp end t)
      (setq n (1+ n)))
    n))


;;; Announce

(provide 'string-recode)


;;; string-recode.el ends here
